import sys
from typing import Optional

from textual import work
from textual.app import App

from torrra._types import Provider
from torrra.commands.config import handle_config_command
from torrra.core.exceptions import ConfigError
from torrra.screens.search import SearchScreen
from torrra.screens.welcome import WelcomeScreen
from torrra.utils.cli import parse_cli_args
from torrra.utils.fs import get_resource_path
from torrra.utils.provider import load_provider, load_provider_from_args


class TorrraApp(App):
    TITLE = "torrra"
    CSS_PATH = get_resource_path("app.css")
    BINDINGS = [
        ("q", "quit", "Quit"),
        ("d", "toggle_dark_mode", "Toggle dark mode"),
        ("escape", "clear_focus", "Clear focus"),
    ]
    ENABLE_COMMAND_PALETTE = False

    def __init__(self, provider: Optional[Provider]) -> None:
        super().__init__()
        self.provider = provider

    @work
    async def on_mount(self) -> None:
        self.theme = "catppuccin-mocha"

        if query := await self.push_screen_wait(WelcomeScreen(provider=self.provider)):
            search_screen = SearchScreen(provider=self.provider, initial_query=query)
            await self.push_screen(search_screen)

    def action_toggle_dark_mode(self) -> None:
        self.theme = (
            "catppuccin-mocha"
            if self.theme == "catppuccin-latte"
            else "catppuccin-latte"
        )

    def action_clear_focus(self) -> None:
        self.set_focus(None)


def main():
    args = parse_cli_args()
    provider = None

    try:
        if args.jackett is not None:
            if len(args.jackett) == 0:
                provider = load_provider("jackett")
            elif len(args.jackett) == 2:
                url, api_key = args.jackett
                provider = load_provider_from_args("jackett", url, api_key)
            else:
                print(
                    "[error] --jackett expects either no args or exactly 2: URL API_KEY"
                )
                sys.exit(1)
        elif args.command == "config":
            handle_config_command(args)
            sys.exit()

        if not provider:
            print(
                "[error] no provider specified\n"
                "run torrra --help for more information",
            )
            sys.exit(1)

        app = TorrraApp(provider=provider)
        app.run()

    except (FileNotFoundError, RuntimeError, ConfigError) as e:
        print(str(e))
        sys.exit(1)


if __name__ == "__main__":
    main()



import argparse

from torrra._version import __version__


def parse_cli_args():
    parser = argparse.ArgumentParser(prog="torrra")
    subparsers = parser.add_subparsers(dest="command")

    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"%(prog)s v{__version__}",
        help="show torrra version and exit",
    )
    parser.add_argument(
        "-j",
        "--jackett",
        nargs="*",
        metavar=("URL", "API_KEY"),
        help="use Jackett provider and/or provide URL and API key",
    )

    # "config" sub-command
    config_parser = subparsers.add_parser("config", help="configure torrra")
    config_parser.add_argument("-g", "--get", metavar="KEY", help="get a config value")
    config_parser.add_argument(
        "-s",
        "--set",
        nargs=2,
        metavar=("KEY", "VALUE"),
        help="set a config key-value pair",
    )
    config_parser.add_argument(
        "-l", "--list", action="store_true", help="list all configs"
    )

    return parser.parse_args()



import asyncio
import json
from pathlib import Path
from typing import Dict, Literal

from platformdirs import site_config_dir, user_config_dir

from torrra._types import Provider
from torrra.providers.jackett import JackettClient


def load_provider(provider: Literal["jackett"]) -> Provider:
    if provider == "jackett":
        return load_jackett_config()


def load_provider_from_args(
    name: Literal["jackett"], url: str, api_key: str
) -> Provider:
    if name == "jackett":
        jc = JackettClient(url=url, api_key=api_key)
        if asyncio.run(jc.validate()):
            return Provider(name="Jackett", url=url, api_key=api_key)
        else:
            raise SystemExit(1)


def load_jackett_config() -> Provider:
    config_path = _find_jackett_config_path()
    config = _load_json_config(config_path)

    port = config.get("Port", 9117)
    host = config.get("LocalBindAddress", "127.0.0.1")
    url = f"http://{host}:{port}"

    try:
        api_key = config["APIKey"]
    except KeyError:
        raise RuntimeError(f"[error] 'APIKey' not found in config: {config_path}")

    jc = JackettClient(url=url, api_key=api_key)
    if asyncio.run(jc.validate()):
        return Provider(name="Jackett", url=url, api_key=api_key)
    else:
        raise SystemExit(1)


def _find_jackett_config_path() -> Path:
    possible_paths = [
        Path(user_config_dir(None)) / "Jackett" / "ServerConfig.json",
        Path(site_config_dir(None)) / "Jackett" / "ServerConfig.json",
    ]

    for path in possible_paths:
        if path.exists():
            return path

    raise FileNotFoundError(
        "[error] jackett config file not found in known locations.\n"
        "checked:\n "
        + "\n ".join(str(p) for p in possible_paths)
        + "\nplease ensure jackett is installed and has run at least once."
    )


def _load_json_config(path: Path) -> Dict:
    try:
        with path.open("r", encoding="utf-8") as f:
            return json.load(f)
    except json.JSONDecodeError as e:
        raise RuntimeError(f"[error] invalid json in config file: {path}\n{e}")
    except Exception as e:
        raise RuntimeError(f"[error] failed to load config file: {path}\n{e}")



from typing import Optional

from textual import on
from textual.app import ComposeResult
from textual.containers import Container, Grid
from textual.screen import Screen
from textual.widgets import Input, Static

from torrra._types import Provider
from torrra._version import __version__
from torrra.utils.fs import get_resource_path

BANNER = """
▀█▀ █▀█ █▀▄ █▀▄ █▀▄ █▀█
 █  █ █ █▀▄ █▀▄ █▀▄ █▀█
 ▀  ▀▀▀ ▀ ▀ ▀ ▀ ▀ ▀ ▀ ▀
"""


class WelcomeScreen(Screen[str]):
    CSS_PATH = get_resource_path("screens/welcome.css")

    def __init__(self, provider: Optional[Provider]) -> None:
        super().__init__()
        self.provider = provider

    def compose(self) -> ComposeResult:
        with Container():
            yield Static(BANNER, id="banner")
            yield Static(
                "\n".join(
                    [
                        "Find and download torrents right from here.",
                        "Powered by libtorrent and Python ❤️",
                    ]
                ),
                id="subtitle",
            )
            yield Input(placeholder="Search...", id="search")
            provider_name = f" - {self.provider.name}" if self.provider else ""
            yield Static(f"v{__version__}{provider_name}", id="version")
            with Container(id="commands_container"):
                with Grid(id="commands"):
                    yield Static("[commands]", id="title", markup=False)
                    yield Static("[esc]ape focus", markup=False)
                    yield Static("esc", classes="key")
                    yield Static("toggle [d]ark mode", markup=False)
                    yield Static("d", classes="key")
                    yield Static("[q]uit", markup=False)
                    yield Static("q", classes="key")

    @on(Input.Submitted, "#search")
    async def handle_search(self, event: Input.Submitted) -> None:
        if query := event.value:
            self.dismiss(query)



from typing import Any

from torrra.core.context import config


def handle_config_command(args: Any):
    if args.get:
        get_res = config.get(args.get)
        print(get_res)
    elif args.set:
        config.set(args.set[0], args.set[1])
    elif args.list:
        list_res = config.list()
        print("\n".join(list_res))



def human_readable_size(size_bytes: float) -> str:
    for unit in ["B", "KB", "MB", "GB", "TB"]:
        if size_bytes < 1024.0:
            return f"{size_bytes:.2f} {unit}"
        size_bytes /= 1024.0
    return f"{size_bytes:.2f} PB"



from typing import Dict, List

import httpx


class JackettClient:
    def __init__(self, url: str, api_key: str, timeout: int = 10):
        self.url = url.rstrip("/")
        self.api_key = api_key
        self.timeout = timeout

    async def search(self, query: str) -> List[Dict]:
        endpoint = f"{self.url}/api/v2.0/indexers/all/results"
        params = {"apikey": self.api_key, "query": query}

        async with httpx.AsyncClient(timeout=self.timeout) as client:
            resp = await client.get(endpoint, params=params)
            resp.raise_for_status()
            return resp.json().get("Results", [])

    async def validate(self) -> bool:
        url = f"{self.url}/api/v2.0/indexers/nonexistent_indexer/results"
        params = {"apikey": self.api_key}

        async with httpx.AsyncClient(timeout=self.timeout) as client:
            try:
                resp = await client.get(url, params=params)
                resp.raise_for_status()
                return True

            except httpx.RequestError:
                print(f"[error] could not connect to jackett at {self.url}.")
                print("please make sure jackett is running and the url is correct.")

            except httpx.HTTPStatusError as e:
                status_code = e.response.status_code

                if status_code == 401:
                    print("[error] invalid jackett api key.")
                    print("double-check the api key you provided.")
                elif status_code == 500 and "nonexistent_indexer" in e.response.text:
                    return True
                else:
                    print(f"[error] jackett returned http {status_code}.")
                    print("unexpected response from jackett. please verify your setup.")

            return False



import time
from typing import List, Optional

import libtorrent as lt
from textual import on, work
from textual.app import ComposeResult
from textual.containers import Container, Horizontal, Vertical
from textual.message import Message
from textual.screen import Screen
from textual.widgets import DataTable, Input, LoadingIndicator, ProgressBar, Static
from textual.widgets.data_table import ColumnKey

from torrra._types import Provider
from torrra.core.context import config
from torrra.providers.jackett import JackettClient
from torrra.utils.fs import get_resource_path
from torrra.utils.helpers import human_readable_size


class SearchScreen(Screen):
    CSS_PATH = get_resource_path("screens/search.css")
    # https://github.com/edward-jazzhands/eds-sandbox/blob/main/python/textual/examples/datatable_expandcol.py
    no_col_width = 3
    title_col_minimum = 25  # dynamic
    size_col_width = 10
    seeders_col_width = 7
    leechers_col_width = 5
    tracker_col_width = 10
    other_cols_total = (
        no_col_width
        + size_col_width
        + seeders_col_width
        + leechers_col_width
        + tracker_col_width
    )

    class SearchResults(Message):
        def __init__(self, results: List, query: str) -> None:
            self.results = results
            self.query = query
            super().__init__()

    def __init__(self, provider: Optional[Provider], initial_query: str):
        super().__init__()
        self.provider = provider
        self.initial_query = initial_query
        # libtorrent
        self.lt_session = None
        self.lt_handle = None
        self.lt_paused = False

    def compose(self) -> ComposeResult:
        search_input = Input(
            placeholder="Search...", id="search", value=self.initial_query
        )
        search_input.border_title = "[$secondary]s[/]earch"
        results_table = DataTable(
            id="results_table",
            cursor_type="row",
            show_cursor=True,
            cell_padding=2,
            classes="hidden",
        )
        results_table.border_title = "[$secondary]r[/]esults"

        with Vertical():
            yield search_input
            with Vertical(id="loader"):
                yield Static()
                yield LoadingIndicator()
            yield results_table
            with Container(id="downloads_container") as c:
                c.border_title = "[$secondary]d[/]ownloads"
                c.can_focus = True
                yield Static("No active downloads")
                with Horizontal(id="bar-and-actions", classes="hidden"):
                    yield ProgressBar(total=100)
                    yield Static(
                        "[$secondary-muted][$secondary]p[/$secondary]ause [$secondary]r[/$secondary]esume",
                        id="actions",
                        disabled=True,
                    )

    def on_mount(self) -> None:
        self.post_message(
            Input.Submitted(self.query_one("#search", Input), self.initial_query)
        )

        table = self.query_one("#results_table", DataTable)
        table.add_column("No.", width=self.no_col_width, key="no_col")
        table.add_column("Title", width=self.title_col_minimum, key="title_col")
        table.add_column("Size", width=self.size_col_width, key="size_col")
        table.add_column("Seeders", width=self.seeders_col_width, key="seeders_col")
        table.add_column("Peers", width=self.leechers_col_width, key="leechers_col")
        table.add_column("Tracker", width=self.tracker_col_width, key="tracker_col")

    def on_unmount(self) -> None:
        # clean libtorrent session
        if self.lt_session and self.lt_handle:
            self.lt_session.remove_torrent(self.lt_handle)

    def on_resize(self) -> None:
        table = self.query_one("#results_table", DataTable)

        total_cell_padding = table.cell_padding * 2 * len(table.columns)
        # space taken for border and padding
        border_and_padding = 4
        available_width = (
            self.size.width
            - self.other_cols_total
            - total_cell_padding
            - border_and_padding
        )

        second_col_width = max(self.title_col_minimum, available_width)
        table.columns[ColumnKey("title_col")].width = second_col_width

    def key_s(self) -> None:
        self.query_one("#search", Input).focus()

    def key_p(self) -> None:
        downloads_container = self.query_one("#downloads_container", Container)
        if not downloads_container.has_focus or not self.lt_handle or self.lt_paused:
            return

        self.lt_handle.pause()
        self.lt_paused = True

    def key_r(self) -> None:
        downloads_container = self.query_one("#downloads_container", Container)
        if (
            not downloads_container.has_focus
            or not self.lt_handle
            or not self.lt_paused
        ):
            return

        self.lt_handle.resume()
        self.lt_paused = False

    @on(Input.Submitted, "#search")
    async def handle_search(self, event: Input.Submitted) -> None:
        query = event.value
        if not query:
            return

        table = self.query_one("#results_table", DataTable)
        loader = self.query_one("#loader", Vertical)
        loader_text = self.query_one("#loader Static", Static)

        table.add_class("hidden")
        table.clear()
        loader.remove_class("hidden")
        loader_text.update(f'Searching for: "[b]{query}[/]"')

        self._perform_search(query)

    @on(DataTable.RowSelected, "#results_table")
    def handle_select(self, event: DataTable.RowSelected) -> None:
        row_key = event.row_key
        magnet_uri = row_key.value

        self.query_one("#bar-and-actions", Horizontal).remove_class("hidden")
        self.query_one("#search", Input).disabled = True
        event.control.disabled = True
        self._download_torrent(magnet_uri)

    @work(exclusive=True, thread=True)
    async def _perform_search(self, query: str) -> None:
        client = self._get_client()
        results = []
        if client:
            try:
                results = await client.search(query)
            except Exception as e:
                self.log.error(f"error during search: {e}")

        self.post_message(self.SearchResults(results, query))

    @on(SearchResults)
    def _show_search_results(self, message: SearchResults) -> None:
        table = self.query_one("#results_table", DataTable)
        loader = self.query_one("#loader", Vertical)
        loader_text = loader.query_one(Static)

        if not message.results:
            loader_text.update(f'Nothing found for "[b]{message.query}[/b]"')
            return

        loader.add_class("hidden")
        table.remove_class("hidden")

        for idx, torrent in enumerate(message.results):
            table.add_row(
                idx + 1,
                torrent["Title"],
                human_readable_size(torrent["Size"]),
                torrent["Seeders"],
                torrent["Peers"],
                torrent["Tracker"],
                key=torrent["MagnetUri"],
            )

        table.focus()

    @work(exclusive=True, thread=True)
    async def _download_torrent(self, magnet_uri: str) -> None:
        self.lt_session = lt.session()  # pyright: ignore
        self.lt_session.listen_on(6881, 6891)

        params = {
            "save_path": config.get("general.download_path"),
            "storage_mode": lt.storage_mode_t.storage_mode_sparse,  # pyright: ignore
        }

        self.lt_handle = lt.add_magnet_uri(  # pyright: ignore
            self.lt_session, magnet_uri, params
        )

        self.app.call_from_thread(
            self._update_download_ui, "[b $success]Fetching Metadata...[/]", 0
        )

        while not self.lt_handle.has_metadata():
            time.sleep(0.5)

        torrent_info = self.lt_handle.get_torrent_info()
        title = torrent_info.name()
        total_size = human_readable_size(torrent_info.total_size())

        status_text_template = (
            f"[b $secondary]Title: [$primary]{title}[/$primary] - "
            "Mode: [$success]{status}[/$success] - "
            "Seeds: {seeds} - "
            "Peers: {peers} - "
        )

        while not self.lt_handle.is_seed():
            s = self.lt_handle.status()
            seed_status = (
                status_text_template.format(
                    status="Paused" if self.lt_paused else "Download",
                    seeds=s.num_seeds,
                    peers=s.num_peers,
                )
                + f"Size: {total_size}[/]"
            )

            self.app.call_from_thread(
                self._update_download_ui,
                seed_status,
                s.progress * 100,
            )

            time.sleep(1)

        while True:
            s = self.lt_handle.status()
            seed_status = (
                status_text_template.format(
                    status="Paused" if self.lt_paused else "Seed",
                    seeds=s.num_seeds,
                    peers=s.num_peers,
                )
                + f"Uploaded: {human_readable_size(s.total_upload)}[/]"
            )

            self.app.call_from_thread(self._update_download_ui, seed_status, 100)
            time.sleep(5)

    def _update_download_ui(self, status: str, progress: float) -> None:
        self.query_one("#downloads_container Static", Static).update(status)
        self.query_one("#downloads_container ProgressBar", ProgressBar).update(
            progress=progress
        )

    def _get_client(self) -> Optional[JackettClient]:
        if not self.provider:
            return

        if self.provider.name == "Jackett":
            return JackettClient(url=self.provider.url, api_key=self.provider.api_key)



from dataclasses import dataclass
from typing import Literal


@dataclass
class Torrent:
    title: str
    magnet_uri: str


@dataclass
class Provider:
    name: Literal["Jackett"]
    url: str
    api_key: str



import os
import sys


def get_resource_path(relative_path: str):
    if hasattr(sys, "_MEIPASS"):
        # running in a pyinstaller bundle
        # resources are typically placed under _MEIPASS/torrra/
        return os.path.join(sys._MEIPASS, "torrra", relative_path)  # pyright: ignore
    else:
        # running in a dev environment
        package_root_dir = os.path.dirname(os.path.dirname(__file__))
        return os.path.join(package_root_dir, relative_path)



import ast
import tomllib
from pathlib import Path
from typing import Any, Dict, List

import tomli_w
from platformdirs import user_config_dir, user_downloads_dir

from torrra.core.exceptions import ConfigError

CONFIG_DIR = Path(user_config_dir("torrra"))
CONFIG_FILE = CONFIG_DIR / "config.toml"


class Config:
    def __init__(self) -> None:
        self.config: Dict[str, Any] = {}
        self._load_config()

    def get(self, key_path: str) -> Any:
        keys = key_path.split(".")
        current = self.config

        try:
            for key in keys:
                current = current[key]
        except (KeyError, TypeError):
            if len(keys) > 1:
                raise ConfigError(f"[error] key does not contain a section: {key_path}")
            else:
                raise ConfigError(f"[error] key not found: {key_path}")

        if isinstance(current, dict):
            raise ConfigError(
                f"[error] key does not contain a value (it's a section): {key_path}"
            )

        return current

    def set(self, key_path: str, value: Any) -> None:
        current = self.config
        keys = key_path.split(".")

        try:
            for key in keys[:-1]:
                if key not in current:
                    current[key] = {}
                elif not isinstance(current[key], dict):
                    raise ConfigError(
                        f"[error] cannot set '{key_path}': '{key}' is not a section"
                    )
                current = current[key]

            try:
                value = ast.literal_eval(value)
            except (ValueError, SyntaxError):
                pass

            current[keys[-1]] = value
            self._save_config()

        except (KeyError, TypeError) as e:
            raise ConfigError(f"[error] failed to set '{key_path}': {str(e)}")

    def list(self) -> List[str]:
        results = []
        for section in self.config:
            for key, value in self.config[section].items():
                if isinstance(value, bool):
                    value = str(value).lower()
                results.append(f"{section}.{key}={value}")

        return results

    def _load_config(self) -> None:
        if not CONFIG_FILE.exists():
            self._create_default_config()
            self._save_config()

        try:
            with open(CONFIG_FILE, "rb") as f:
                self.config = tomllib.load(f)
        except Exception as e:
            print(f"[error] loading config: {e}")

    def _create_default_config(self) -> None:
        self.config = {
            "general": {
                "download_path": user_downloads_dir(),
                "remember_last_path": True,
            }
        }

    def _save_config(self) -> None:
        CONFIG_DIR.mkdir(parents=True, exist_ok=True)
        with open(CONFIG_FILE, "wb") as f:
            tomli_w.dump(self.config, f)
